package net.sarcommand.fft;

import java.io.*;
import java.util.Arrays;

/**
 * @author A. O. Ergueven, Torsten Heup<br>
 * Represents a complex signal of any kind. This class holds onedimensional signals as well as twodimensional
 * signals. The complex numbers are represented by two float arrays for the real and the imaginary part.
 */
public class ComplexSignal implements Externalizable {
    /**
     * Holds the real parts of the complex numbers in this signal.
     */
    protected float[] _real;

    /**
     * Holds the imaginary parts of the complex numbers in this signal.
     */
    protected float[] _imag;

    /**
     * This signal's total length.
     */
    public int length;

    /**
     * If the signal is twodimensional, this variable holds the signals width.
     */
    public int width;

    /**
     * If the signal is twodimensional, this variable holds the signals height.
     */
    public int height;

    /**
     * Creates a new complex signal. This constructor is meant to be used only be the externalization
     * mechanisms and should not be invoked.
     */
    public ComplexSignal(){
    }

    /**
     * Creates a new onedimensional signal with the length 'size'.
     * @param size Length of signal
     */
    public ComplexSignal(final int size) {
        _real = new float[size];
        _imag = new float[size];
        width = size;
        height = 0;
        length = width;
    }

    /**
     * Creates a new twodimensional signal.
     *
     * @param width Signal's width
     * @param height Signal's height
     */
    public ComplexSignal(final int width, final int height){
        this.width = width;
        this.height = height;
        length = width * height;

        _real = new float[length];
        _imag = new float[length];
    }

    /**
     * Returns the signal's real part.
     *
     * @return real part
     */
    public float[] getRealPart(){
        return _real;
    }

    /**
     * Returns the signal's imaginary part.
     * @return imaginary part
     */
    public float[] getImaginaryPart(){
        return _imag;
    }

    /**
     * Creates an exact copy of this signal.
     *
     * @return Signal's clone
     */
    public ComplexSignal copy(){
        final ComplexSignal signal;
        if (is2D())
            signal = new ComplexSignal(width, height);
        else
            signal = new ComplexSignal(length);
        System.arraycopy(_real,0,signal._real,0,_real.length);
        System.arraycopy(_imag,0,signal._imag,0,_imag.length);
        return signal;
    }

    /**
     * Returns whether this signal is twodimensional.
     *
     * @return true if signal is 2D, false otherwise.
     */
    public boolean is2D(){
        return width != 0 && height != 0;
    }

    /**
     * @see Externalizable#writeExternal(java.io.ObjectOutput)
     */
    public void writeExternal(ObjectOutput out) throws IOException {
        out.writeBoolean(is2D());

        out.writeInt(length);
        out.writeInt(width);
        out.writeInt(height);

        out.writeObject(_real);
        out.writeObject(_imag);

        out.flush();
    }

    /**
     * @see Externalizable#readExternal(java.io.ObjectInput)
     */
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        final boolean is2D = in.readBoolean();

        length = in.readInt();
        width = in.readInt();
        height = in.readInt();

        _real = (float[]) in.readObject();
        _imag = (float[]) in.readObject();
    }

    /**
     * Returns whether this signal is completely equal to the specified one.
     *
     * @param o Signal to compare with
     * @return true if signals are equal, false otherwise
     */
    public boolean equals(final Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        final ComplexSignal that = (ComplexSignal) o;

        if (height != that.height) return false;
        if (length != that.length) return false;
        if (width != that.width) return false;
        if (!Arrays.equals(_imag, that._imag)) return false;
        return Arrays.equals(_real, that._real);
    }
}
