package net.sarcommand.fft;

import net.sarcommand.util.ProgressNotifier;

/**
 * @author A. O. Ergueven, Torsten Heup<br>
 * Implements a naive onedimensional discrete fourier transformation.
 */
public class DFT1 extends FourierTransform1D{
    /**
     * Matrix used to cache the cosinus parts for the roots of unity used during calculation.
     */
    float[] _cosinusMatrix;

    /**
     * Matrix used to cache the sinus parts for the roots of unity used during calculation.
     */
    float[] _sinusMatrix;

    /**
     * Creates a new instance of this DFT.
     */
    public DFT1() {
        _cosinusMatrix = new float[0];
        _sinusMatrix = new float[0];
    }

    /**
     * Creates the fourier matrix used for transformation.
     *
     * @param width This signal's length.
     */
    protected void buildVandermondeMatrix(final int width){
        _cosinusMatrix = new float[width];
        for (int i = 0; i < width; i++) {
            _cosinusMatrix[i] = (float) Math.cos(2 * Math.PI * i / width);
        }

        _sinusMatrix = new float[width];
        for (int i = 0; i < width; i++) {
            _sinusMatrix[i] = (float) Math.sin(2 * Math.PI * i / width);
        }
    }

    /**
     * Applies a forward dft to the signal.
     *
     * @param signal Signal to transform.
     * @param notifier Notifier
     */
    public void transform(final ComplexSignal signal, final ProgressNotifier notifier) {
        if (signal.length != _cosinusMatrix.length)
            buildVandermondeMatrix(signal.length);

        final ComplexSignal copy = signal.copy();

        final float[] real = copy.getRealPart();
        final float[] imag = copy.getImaginaryPart();
        final float[] destReal = signal.getRealPart();
        final float[] destImag = signal.getImaginaryPart();

        int m = signal.length;
        for (int u = 0; u < m; u++) {
            float sumRe = 0;
            float sumIm = 0;
            for (int c = 0; c < m; c++) {
                float gRe = real[c];
                float gIm = imag[c];
                int k = (u * c) % m;
                float cosPhi = _cosinusMatrix[k];
                float sinPhi = _sinusMatrix[k];
                //complex multiplication: (gRe + i gIm) * (cosPhi + i sinPhi)
                sumRe += gRe * cosPhi - gIm * sinPhi;
                sumIm += gRe * sinPhi + gIm * cosPhi;
            }
            destReal[u] = sumRe;
            destImag[u] = sumIm;
            if (notifier != null)
                notifier.setProgress("applying dft",u/m);
        }
    }

    /**
     * Applies an inverse dft to the specified signal.
     *
     * @param signal Signal to transform
     * @param notifier Notifier
     */
    public void inverseTransform(final ComplexSignal signal, final ProgressNotifier notifier) {
        if (signal.length != _cosinusMatrix.length)
            buildVandermondeMatrix(signal.length);

        final ComplexSignal copy = signal.copy();

        final float[] real = copy.getRealPart();
        final float[] imag = copy.getImaginaryPart();
        final float[] destReal = signal.getRealPart();
        final float[] destImag = signal.getImaginaryPart();

        int m = signal.length;
        for (int u = 0; u < m; u++) {
            float sumRe = 0;
            float sumIm = 0;
            for (int c = 0; c < m; c++) {
                float gRe = real[c] / m;
                float gIm = imag[c] / m;
                int k = (u * c) % m;
                float cosPhi = _cosinusMatrix[k];
                float sinPhi = - _sinusMatrix[k];
                sumRe += gRe * cosPhi - gIm * sinPhi;
                sumIm += gRe * sinPhi + gIm * cosPhi;
            }
            destReal[u] = sumRe;
            destImag[u] = sumIm;
            if (notifier != null)
                notifier.setProgress("applying inverse DFT",u/m);
        }
    }
}
