package net.sarcommand.fft;

import net.sarcommand.util.ProgressNotifier;

/**
 * @author A. O. Ergueven, Torsten Heup<br>
 * Implements a simple twodimensional discrete fourier transformation. The DFT1 class will be used to transform
 * all rows and columns when applying the twodimensional transform.
 */
public class DFT2 extends FourierTransform2D{

    /**
     * Instanciates a new twodimensional dft.
     */
    public DFT2() {
    }

    /**
     * Apply the fourier transform to the given input data.
     *
     * @param inputData Signal to transform
     * @param notifier Notifier
     */
    public void transform(final ComplexSignal inputData, final ProgressNotifier notifier) {
        if (!inputData.is2D())
            throw new IllegalArgumentException("Expected a two-dimensional signal.");

        final int width = inputData.width;
        final int height = inputData.height;

        image = inputData;

        DFT1 dft = new DFT1();
        ComplexSignal signal = new ComplexSignal(width);

        for (int v = 0; v < height; v++) {
            getRow(v, signal);
            dft.transform(signal, null);
            putRow(v, signal);
            if (notifier != null)
                notifier.setProgress("applying dft",((float)v)/height/2);
        }

        if (height != width) {
            dft = new DFT1();
            signal = new ComplexSignal(height);
        }

        for (int u = 0; u < width; u++) {
            getCol(u, signal);
            dft.transform(signal, null);
            putCol(u, signal);
            if (notifier != null)
                notifier.setProgress("applying dft",0.5f + ((float)u)/width/2);
        }
    }

    /**
     * Apply the inverse fourier transform to the given input data.
     *
     * @param inputData Signal to transform
     * @param notifier Notifier
     */
    public void inverseTransform(final ComplexSignal inputData, final ProgressNotifier notifier) {
        if (!inputData.is2D())
            throw new IllegalArgumentException("Expected a two-dimensional signal.");

        final int width = inputData.width;
        final int height = inputData.height;

        image = inputData;

        DFT1 dft = new DFT1();
        ComplexSignal signal = new ComplexSignal(inputData.width);

        for (int v = 0; v < height; v++) {
            getRow(v, signal);
            dft.inverseTransform(signal, null);
            putRow(v, signal);
            if (notifier != null)
                notifier.setProgress("applying inverse dft",((float)v)/height/2);
        }

        if (height != width) {
            dft = new DFT1();
            signal = new ComplexSignal(height);
        }

        for (int u = 0; u < width; u++) {
            getCol(u, signal);
            dft.inverseTransform(signal, null);
            putCol(u, signal);
            if (notifier != null)
                notifier.setProgress("applying inverse dft",0.5f+((float)u)/width/2);
        }
    }
}
