package net.sarcommand.fft;

import net.sarcommand.util.ProgressNotifier;


/**
 * @author A. O. Ergueven, Torsten Heup<br>
 * A fast fourier transform implementation following the Cooley-Tukey
 * algorithm (Radix-2). This approach offers the fastest transform, however
 * it works for power of two width signals only. One can transform signals
 * of any width by zero padding them to the next power of two.<hr>
 */
public class FFT1 extends FourierTransform1D {
    /**
     * This field caches the roots of unity which occur periodically for
     * this evaluation. This way multiple transforms on same-width signals
     * can be computed faster.
     */
    private final int bandWidth;
    /**
     * Creates a new 1-dimensional fast fourier transform for signals of
     * the specified band width.
     *
     * @param bandWidth Length of a signal band.
     */
    public FFT1(final int bandWidth) {
        this.bandWidth = bandWidth;
        final double v = (Math.log(bandWidth) / Math.log(2.));
        if (v % 1 != 0)
            throw new IllegalArgumentException("This implemetation of the fast" +
                "fourier transform expects a power of twolwngth signal.");
    }

    /**
     * Transforms the given vector using the Cooley-Tukey fast fourier
     * transform algorithm.
     *
     */
    public void transform(final ComplexSignal signal, final ProgressNotifier notifier) {
        final float[] unityRoot = computeUnityRoot(bandWidth);
        fft(signal.getRealPart(), signal.getImaginaryPart(), signal.width, 0, unityRoot);
    }

    /**
     * Applies the inverse fourier transform to the given signal using the
     * Cooley-Tukey fast fourier transform algorithm.
     *
     * @param signal Complex signal to transform.
     */
    public void inverseTransform(final ComplexSignal signal, final ProgressNotifier notifier) {
        final float[] real = signal.getRealPart();
        final float[] imag = signal.getImaginaryPart();

        final float dividerReal = 1.f / signal.length;

        for (int i = 0; i < signal.length; i++) {
            real[i] *= dividerReal;
            imag[i] *= dividerReal;
        }

        final float[] wInverse = computeUnityRoot(signal.length);
        wInverse[1]*= -1;
        fft(real, imag, signal.length, 0, wInverse);
    }

    /**
     * Recursively transforms the given complex vector using the
     * Cooley-Tukey algorithm.
     *
     * @param n  Numer of elements to transform.
     * @param lo Beginindex at which to start transformation.
     */
    private void fft(final float[] aReal, final float[] aImag, int n, int lo,
                     final float[] w) {
        int i, m;
        float zReal;
        float zImag;
        float hReal;
        float hImag;
        float vReal = w[0];
        float vImag = w[1];

        if (n > 1) {
            m = n / 2;
            zReal = 1;
            zImag = 0;
            for (i = lo; i < lo + m; i++) {
                /* h = a[i] - a[i+m] */
                hReal = aReal[i] - aReal[i+m];
                hImag = aImag[i] - aImag[i+m];

                /* a[i] = a[i] + a[i+m] */
                aReal[i] = aReal[i] + aReal[i+m];
                aImag[i] = aImag[i] + aImag[i+m];

                /* a[i+m] = h * z */
                aReal[i+m] = hReal * zReal - hImag * zImag;
                aImag[i+m] = hReal * zImag + hImag * zReal;

                /* z = z * w */
                float temp = zReal * vReal - zImag * vImag;
                zImag = zReal * vImag + zImag * vReal;
                zReal = temp;
            }

            /* v = w� */
            float[] v = new float[2];
            v[0] = vReal * vReal - vImag * vImag;
            v[1] = vReal * vImag + vImag * vReal;

            fft(aReal, aImag, m, lo, v);
            fft(aReal, aImag, m, lo + m, v);

            merge(aReal, aImag, n, lo);
        }
    }

    /**
     * Merges the two previously transformed parts of the given vector
     * together.
     *
     * @param n  Number of elements to merge.
     * @param lo Beginindex at which to start the merge.
     */
    private void merge(final float[] aReal, final float[] aImag, final int n, final int lo) {
        int i, m = n / 2;
        final float[] bReal = new float[m];
        final float[] bImag = new float[m];

        for (i = 0; i < m; i++) {
            bReal[i] = aReal[lo + i];
            bImag[i] = aImag[lo + i];
        }
        for (i = 0; i < m; i++) {
            aReal[lo + i + i + 1] = aReal[lo + i + m];
            aImag[lo + i + i + 1] = aImag[lo + i + m];
        }
        for (i = 0; i < m; i++){
            aReal[lo + i + i] = bReal[i];
            aImag[lo + i + i] = bImag[i];
        }
    }
}
