package net.sarcommand.fft;

import net.sarcommand.util.ProgressNotifier;

/**
 * @author A. O. Ergueven, Torsten Heup<br>
 * This class implements a 2-dimensional fast fourier transform using the
 * Cooley-Tukey radix2 algorithm. This algorithm is on average the fastest of
 * the fourier transforms in this package, however, it only computes for
 * vectors with a power of 2 width. Therefore, the data to be transformed will
 * be zero padded. This implementation determines the 2-dimensional FFT by
 * computing a 1-dimensional FFT for each row and then again for each column.
 */
public class FFT2 extends FourierTransform2D {
    /**
     * Creates a new FFT2 instance.
     */
    public FFT2() {
    }

    /**
     * Apply the fourier transform to the given input data.
     *
     * @param inputData float[] containing the input pixels.
     */
    public void transform(final ComplexSignal inputData, final ProgressNotifier notifier) {
        if (!inputData.is2D())
            throw new IllegalArgumentException("Expected a two-dimensional signal.");

        final int width = inputData.width;
        final int height = inputData.height;

        if (!(isPowerOf2(width) && isPowerOf2(height)))
            throw new IllegalArgumentException("This implementation only supports power of two sized images.");

        image = inputData;

        FFT1 fft = new FFT1(width);
        ComplexSignal signal = new ComplexSignal(width);

        for (int v = 0; v < height; v++) {
            getRow(v, signal);
            fft.transform(signal, null);
            putRow(v, signal);
            if (notifier != null)
                notifier.setProgress("applying fft",((float)v)/height/2);
        }

        if (height != width) {
            fft = new FFT1(height);
            signal = new ComplexSignal(height);
        }

        for (int u = 0; u < width; u++) {
            getCol(u, signal);
            fft.transform(signal, null);
            putCol(u, signal);
            if (notifier != null)
                notifier.setProgress("applying fft",0.5f+((float)u)/width/2);
        }
    }

    /**
     * Apply the inverse fourier transform to the given input data.
     *
     * @param inputData float[] containing the input pixels.
     */
    public void inverseTransform(final ComplexSignal inputData, final ProgressNotifier notifier) {
        if (!inputData.is2D())
            throw new IllegalArgumentException("Expected a two-dimensional signal.");

        final int width = inputData.width;
        final int height = inputData.height;

        if (!(isPowerOf2(inputData.width) && isPowerOf2(inputData.height)))
            throw new IllegalArgumentException("This implementation only supports power of two sized images.");

        image = inputData;

        FFT1 fft = new FFT1(width);
        ComplexSignal signal = new ComplexSignal(inputData.width);

        for (int v = 0; v < height; v++) {
            getRow(v, signal);
            fft.inverseTransform(signal, null);
            putRow(v, signal);
            if (notifier != null)
                notifier.setProgress("applying inverse fft",((float)v)/height/2);
        }

        if (height != width) {
            fft = new FFT1(height);
            signal = new ComplexSignal(height);
        }

        for (int u = 0; u < width; u++) {
            getCol(u, signal);
            fft.inverseTransform(signal, null);
            putCol(u, signal);
            if (notifier != null)
                notifier.setProgress("applying inverse fft",0.5f+((float)u)/width/2);
        }
    }

    /**
     * Transforms the given image to a power of two image by padding it
     * with zeros accordingly. The result will be stored in the
     *
     * @param intputData input data
     */
    public ComplexSignal createPowerOf2Image(final ComplexSignal intputData){
        final int origWidth = intputData.width;
        final int origHeight = intputData.height;
        final int newWidth = (int) (Math.pow(2, Math.ceil(Math.log(origWidth) / Math.log(2))));
        final int newHeight = (int) (Math.pow(2, Math.ceil(Math.log(origHeight) / Math.log(2))));

        final ComplexSignal signal = new ComplexSignal(newWidth, newHeight);

        final float[] origReal = intputData.getRealPart();
        final float[] origImag = intputData.getImaginaryPart();
        final float[] destReal = signal.getRealPart();
        final float[] destImag = signal.getImaginaryPart();

        for (int i = 0; i < origHeight; i++) {
            final int offset = i * newWidth;
            for (int j = 0; j < origWidth; j++) {
                destReal[offset + j] = origReal [i * origWidth + j];
                destImag[offset + j] = origImag [i * origWidth + j];
            }
        }

        return signal;
    }
}

