package net.sarcommand.fft;

import net.sarcommand.util.ProgressNotifier;

/**
 * @author A. O. Ergueven, Torsten Heup<br>
 * Abstract superclass for all fourier transformation imeplemtations in this
 * package.
 */
public abstract class FourierTransform {
    /**
     * Returns the nth primitive root of unity.
     */
    public float[] computeUnityRoot(final int n) {
        return new float[]{(float) Math.cos(2 * Math.PI / n),
            (float) Math.sin(2 * Math.PI / n)};
    }

    /**
     * Converts the given frequency image to a byte array by scaling the
     * values to the range of 0 to 255. This method will use the natural
     * logarithm to scale the values.
     *
     * @param frequencyImage FrequencyImage to convert.
     * @return byte[] containing the scaled values.
     */
    public static byte[] convertToBytes(final ComplexSignal frequencyImage){
        final int imageSize = frequencyImage.length;
        final byte[] pixels = new byte[imageSize];
        final float[] real = frequencyImage.getRealPart();
        final float[] imag = frequencyImage.getImaginaryPart();

        float maximumPowerValue = 0f;
        double value;
        for (int i = 0; i < imageSize; i++) {
            value = Math.sqrt(real[i] * real[i] + imag[i] * imag[i]);
            if (value > maximumPowerValue)
                maximumPowerValue = (float) value;
        }

        double max = Math.log(maximumPowerValue + 1.0);
        double scale = 255 / max;

        for (int i = 0; i < pixels.length; i++) {     //ToDo don't compute twice here
            double p = Math.sqrt(real[i] * real[i] + imag[i] * imag[i]);
            if (p < 0) p = -p;
            double plog = Math.log(p + 1.0);
            int pint = (int) (plog * scale);
            pixels[i] = (byte) (0xFF & pint);
        }
        return pixels;
    }

    /**
     * Apply the fourier transform to the given input data.
     *
     * @param inputData float[] containing the input pixels.
     */
    public abstract void transform(final ComplexSignal inputData, final ProgressNotifier notifier);

    /**
     * Apply the inverse fourier transform to the given input data.
     *
     * @param inputData float[] containing the input pixels.
     */
    public abstract void inverseTransform(final ComplexSignal inputData, final ProgressNotifier notifier);
}
