package net.sarcommand.fft;

import net.sarcommand.util.ProgressNotifier;
import org.apache.log4j.Logger;

import java.awt.*;
import java.awt.image.*;

/**
 * @author A. O. Ergueven, Torsten Heup<br>
 * Abstract superclass for all 2-dimensional fourier transforms in this
 * package.
 */
public abstract class FourierTransform2D extends FourierTransform{
    private final static Logger __log = Logger.getLogger(FourierTransform2D.class);

    /**
     * The complex image data.
     */
    protected ComplexSignal image;

    /**
     * Performs a quadrant shift.
     *
     * @param pixels Pixels of the image to shift.
     * @param width The image's width.
     * @param height The image's height.
     */
    public static void shiftQuadrants(final byte[] pixels, final int width, final int height) {
        final int w2 = width / 2;
        final int h2 = height / 2;
        final int h2w = h2 * width;
        for (int i = 0; i < pixels.length / 2 + 1; i++) {
            final int bandOffset = i % width >= w2 ? -w2 : w2;
            if (i / width >= h2) {
                byte cache = pixels[i];
                pixels[i] = pixels[i - h2w + bandOffset];
                pixels[i - h2w + bandOffset] = cache;
            } else {
                byte cache = pixels[i];
                pixels[i] = pixels[i + h2w + bandOffset];
                pixels[i + h2w + bandOffset] = cache;
            }
        }
    }

    /**
     * Creates a grayscale BufferdImage of type {@link BufferedImage#TYPE_BYTE_GRAY}. Therefore,
     * the frequency values will be scaled to a range of 0-255 and a quadrant shift
     * will be applied afterwards.
     *
     * @param frequencyImage Frequency image to use as source.
     * @return Human-friendly grayscale version of the frequency image.
     */
    public static BufferedImage fourierSignalToGrayscale(final ComplexSignal frequencyImage){
        final int width = frequencyImage.width;
        final int height = frequencyImage.height;
        __log.info("Creating "+width+"x"+height+" grayscale image.");

        final byte[] bytes = convertToBytes(frequencyImage);
        shiftQuadrants(bytes, width, height);

        final DataBuffer newBuffer = new DataBufferByte(bytes, bytes.length);
        final Raster newRaster = Raster.createPackedRaster(newBuffer, width, height, 8, new Point());
        final BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        image.setData(newRaster);

        return image;
    }

    /**
     * Creates a grayscale BufferedImage of type {@link BufferedImage#TYPE_BYTE_GRAY} from the given
     * complex signal. It will be assumed that the signal only holds real numbers as components.
     *
     * @param signal Signal to convert
     * @return BufferedImage containing signal
     */
    public static BufferedImage realToGrayscale(final ComplexSignal signal){
        if (!signal.is2D())
            throw new IllegalArgumentException("Expected a two-dimensional signal here!");

        final int width = signal.width;
        final int height = signal.height;
        __log.info("Creating "+width+"x"+height+" grayscale image.");

        final byte[] bytes = new byte[signal.length];
        final float[] real = signal.getRealPart();

        for (int i = 0; i < signal.length; i++) {
            bytes[i] = (byte) real[i];   //ToDo scale
        }

        final DataBuffer newBuffer = new DataBufferByte(bytes, bytes.length);
        final Raster newRaster = Raster.createPackedRaster(newBuffer, width, height, 8, new Point());
        final BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        image.setData(newRaster);

        return image;
    }

    /**
     * Converts the pixels of an image to a complex signal.
     *
     * @param image Image to convert.
     * @return ComplexSignal containing the pixel values.
     * @throws IllegalArgumentException If an unknown image type was specified.
     */
    public static ComplexSignal imageToSignal(final BufferedImage image) throws IllegalArgumentException{
        final WritableRaster raster = image.getRaster();
        final DataBuffer db = raster.getDataBuffer();

        switch (db.getDataType()) {
            case DataBuffer.TYPE_BYTE:
                final DataBufferByte buffer = (DataBufferByte) db;
                final byte[] data = buffer.getData();

                final ComplexSignal signal = new ComplexSignal(image.getWidth(), image.getHeight());
                final float[] pixels = signal.getRealPart();

                for (int i = 0; i < data.length; i++)
                    pixels[i] = data[i];

                return signal;
            default:
                throw new IllegalArgumentException("Unsupported image type");
        }
    }

    /**
     * Transforms this BufferedImage to a frequencyImage.
     * @param image Image to transform.
     * @return float[] containing the frequency image.
     */
    public ComplexSignal transform(final BufferedImage image, final ProgressNotifier notifier) throws IllegalArgumentException{
        final ComplexSignal pixels = imageToSignal(image);
        transform(pixels, notifier);
        return pixels;
    }

    /**
     * Applies an inverse fourier transform to this BufferedImage.
     *
     * @param image Image to transform.
     * @return float[] containing the frequency image.
     */
    public ComplexSignal inverseTransform(final BufferedImage image, final ProgressNotifier notifier) throws IllegalArgumentException{
        final ComplexSignal pixels = imageToSignal(image);
        inverseTransform(pixels, notifier);
        return pixels;
    }

    /**
     * Returns row number 'v' from the complex image being computed
     * and stores it's contents in 'rowC'.
     *
     * @param v Index of row to retrieve.
     */
    protected void getRow(int v, ComplexSignal signal) {
        final float[] imgReal = image.getRealPart();
        final float[] imgImag = image.getImaginaryPart();
        final float[] sigReal = signal.getRealPart();
        final float[] sigImag = signal.getImaginaryPart();

        int i = v * image.width; //start index of row v
        for (int u = 0; u < image.width; u++) {
            sigReal[u] = imgReal[i + u];
            sigImag[u] = imgImag[i + u];
        }
    }

    /**
     * Stores the contents of the complex vector 'rowC' in row number 'v'
     * of the complex image being computed.
     *
     * @param v Index of row to store.
     */
    protected void putRow(int v, ComplexSignal signal) {
        final float[] imgReal = image.getRealPart();
        final float[] imgImag = image.getImaginaryPart();
        final float[] sigReal = signal.getRealPart();
        final float[] sigImag = signal.getImaginaryPart();

        int i = v * image.width; //start index of row v
        for (int u = 0; u < image.width; u++) {
            imgReal[i + u] = sigReal[u];
            imgImag[i + u] = sigImag[u];
        }
    }

    /**
     * Returns the contents of the column indexed by u in the complex
     * image being computed.
     *
     * @param u Index of column to retrieve.
     */
    protected void getCol(int u, ComplexSignal signal) {
        final float[] imgReal = image.getRealPart();
        final float[] imgImag = image.getImaginaryPart();
        final float[] sigReal = signal.getRealPart();
        final float[] sigImag = signal.getImaginaryPart();

        for (int v = 0; v < image.height; v++) {
            sigReal[v] = imgReal[v * image.width + u];
            sigImag[v] = imgImag[v * image.width + u];
        }
    }

    /**
     * Stores the contents of the complex vector 'colV' in column number 'u'
     * of the complex image being computed.
     *
     * @param u Index of column to store.
     */
    protected void putCol(int u, ComplexSignal signal) {
        final float[] imgReal = image.getRealPart();
        final float[] imgImag = image.getImaginaryPart();
        final float[] sigReal = signal.getRealPart();
        final float[] sigImag = signal.getImaginaryPart();

        for (int v = 0; v < image.height; v++) {
            imgReal[v * image.width + u] = sigReal[v];
            imgImag[v * image.width + u] = sigImag[v];
        }
    }

    /**
     * Returns true if the given int is a power of two value.
     *
     * @param length Signal width to be checked.
     * @return true if width is a power of two, false otherwise.
     */
    public static boolean isPowerOf2(final int length){
        final double d = (Math.log(length) / Math.log(2)) % 1;
        return d == 0;
    }

    /**
     * Returns true if the given int is a power of two value.
     *
     * @return true if width is a power of two, false otherwise.
     */
    public static boolean isPowerOf2(final Dimension size){
        return isPowerOf2(size.width) && isPowerOf2(size.height);
    }
}
