package net.sarcommand.fft.demo;

import net.sarcommand.fft.ComplexSignal;
import net.sarcommand.fft.DFT2;
import net.sarcommand.fft.FFT2;
import net.sarcommand.fft.FourierTransform2D;
import net.sarcommand.swingext.core.imagepanel.JImagePanel;
import net.sarcommand.util.ProgressNotifier;
import org.apache.log4j.Logger;

import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.*;

/**
 * @author A. O. Ergueven, Torsten Heup
 */
public class FFTDemo extends JFrame implements ProgressNotifier {
    private final static Logger __log = Logger.getLogger(FFTDemo.class);

    private JPanel _cPane;
    private JImagePanel _normalImage;
    protected JImagePanel _freqImage;

    private JButton _dftButton;
    private JButton _fftButton;
    private JButton _iDftButton;
    private JButton _iFftButton;

    private JButton _openNormalButton;
    private JButton _openFreqButton;
    private JButton _saveFreqButton;

    private ImageIcon _iconArrowRight;
    private ImageIcon _iconArrowLeft;
    private ImageIcon _iconOpen;
    private ImageIcon _iconSave;

    private JFileChooser _sharedFileChooser;
    private JProgressBar _progressBar;
    protected JDialog _progressDlg;
    protected JTextField _statusTF;

    protected JPanel _glassPane;

    private ComplexSignal _realSignal;
    private ComplexSignal _fourierSignal;
    private FileFilter _imageFilter;
    private FileFilter _complexFilter;

    public FFTDemo() {
        init();
    }

    /**
     * Initializes this component. This method should be invoked by every constructor as it take's care of
     * setting up the required fields and layout definitons.
     */
    protected void init() {
        initSharedFileChooser();
        loadImages();
        initComponents();
        initLayout();
        setupEventHandlers();
        initFrame();
    }


    protected void loadImages() {
        final ClassLoader loader = getClass().getClassLoader();
        try {
            _iconArrowLeft = new ImageIcon(ImageIO.read(loader.getResourceAsStream("icons/arrowLeft_22x22.png")));
            _iconArrowRight = new ImageIcon(ImageIO.read(loader.getResourceAsStream("icons/arrowRight_22x22.png")));
            _iconOpen = new ImageIcon(ImageIO.read(loader.getResourceAsStream("icons/open_22x22.png")));
            _iconSave = new ImageIcon(ImageIO.read(loader.getResourceAsStream("icons/save_22x22.png")));
        } catch (IOException e) {
            System.err.println("Missing resources...");
            e.printStackTrace();  //To change body of catch statement use File | Settings | File Templates.
        }
    }

    /**
     * Initializes all graphical subcomponents and defines their behavious by setting the accordant
     * properties.
     */
    protected void initComponents() {
        _cPane = new JPanel();

        _normalImage = new JImagePanel(new BufferedImage(3, 3, BufferedImage.TYPE_INT_ARGB));
        _freqImage = new JImagePanel(new BufferedImage(3, 3, BufferedImage.TYPE_INT_ARGB));

        _normalImage.setScaleRate(3.);
        _freqImage.setScaleRate(3.);

        _normalImage.setBorder(BorderFactory.createLoweredBevelBorder());
        _freqImage.setBorder(BorderFactory.createLoweredBevelBorder());

        _dftButton = new JButton("DFT", _iconArrowRight);
        _dftButton.setHorizontalTextPosition(SwingConstants.LEFT);
        _fftButton = new JButton("FFT2", _iconArrowRight);
        _fftButton.setHorizontalTextPosition(SwingConstants.LEFT);

        _iDftButton = new JButton("DFT", _iconArrowLeft);
        _iFftButton = new JButton("FFT2", _iconArrowLeft);

        _openNormalButton = new JButton(_iconOpen);
        _openFreqButton = new JButton(_iconOpen);
        _saveFreqButton = new JButton(_iconSave);

        _progressBar = new JProgressBar(JProgressBar.HORIZONTAL);
        _progressBar.setStringPainted(true);
        _progressBar.setMinimum(0);
        _progressBar.setMaximum(100);

        _progressDlg = new JDialog(this, false);
        _progressDlg.setContentPane(_progressBar);
        _progressDlg.pack();
        _progressDlg.setLocationRelativeTo(null);
        _progressDlg.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);

        _statusTF = new JTextField();

        _glassPane = new GrayOutPane();

        updateTransformButtons();
    }

    /**
     * Initializes the layout definitions for this class.
     */
    protected void initLayout() {
        final JPanel buttonPanel = new JPanel();
        buttonPanel.setLayout(new GridLayout(4, 1, 5, 5));
        buttonPanel.add(_dftButton);
        buttonPanel.add(_fftButton);
        buttonPanel.add(_iDftButton);
        buttonPanel.add(_iFftButton);

        final JToolBar leftToolBar = new JToolBar(JToolBar.HORIZONTAL);
        leftToolBar.setFloatable(false);
        leftToolBar.add(_openNormalButton);

        final JToolBar rightToolBar = new JToolBar(JToolBar.HORIZONTAL);
        rightToolBar.setFloatable(false);
        rightToolBar.add(_openFreqButton);
        rightToolBar.add(_saveFreqButton);

        _cPane.setLayout(new GridBagLayout());
        _cPane.add(leftToolBar,
            new GridBagConstraints(0, 0, 1, 1, 0.5, 0.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(8, 8, 8, 8), 0, 0));
        _cPane.add(_normalImage,
            new GridBagConstraints(0, 1, 1, 1, 0.5, 1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(8, 8, 8, 8), 0, 0));
        _cPane.add(buttonPanel,
            new GridBagConstraints(1, 0, 1, 2, 0.0, 1.0, GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(3, 8, 8, 3), 0, 0));
        _cPane.add(rightToolBar,
            new GridBagConstraints(2, 0, 1, 1, 0.5, 0.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(8, 8, 8, 8), 0, 0));
        _cPane.add(_freqImage,
            new GridBagConstraints(2, 1, 1, 1, 0.5, 1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(8, 8, 8, 8), 0, 0));
        _cPane.add(_statusTF,
            new GridBagConstraints(0, 2, 3, 1, 1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(8, 8, 8, 8), 0, 0));
    }

    /**
     * Creates and installs the required event handlers.
     */
    protected void setupEventHandlers() {
        final ActionListener actionListener = new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                final Object source = e.getSource();
                if (source == _openNormalButton)
                    openNormalImage();
                else if (source == _fftButton)
                    invokeFFT();
                else if (source == _iFftButton)
                    invokeInverseFFT();
                else if (source == _dftButton)
                    invokeDFT();
                else if (source == _iDftButton)
                    invokeInverseDFT();
                else if (source == _saveFreqButton)
                    saveFrequencyImage();
                else if (source == _openFreqButton)
                    openFrequencyImage();
            }
        };

        _openNormalButton.addActionListener(actionListener);
        _dftButton.addActionListener(actionListener);
        _fftButton.addActionListener(actionListener);
        _iDftButton.addActionListener(actionListener);
        _iFftButton.addActionListener(actionListener);
        _saveFreqButton.addActionListener(actionListener);
        _openFreqButton.addActionListener(actionListener);
    }

    /**
     * Initializes the frame.
     */
    protected void initFrame() {
        setSize(800, 600);
        setLocationRelativeTo(null);
        setTitle("FFT2 Demo");
        setContentPane(_cPane);
        setGlassPane(_glassPane);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setVisible(true);
    }

    private void invokeInverseFFT(){
        __log.info("Invoking inverse fft");
        final TransformThread t = new TransformThread(this,"Inverse radix2 FFT") {
            public void transform() {
                final FFT2 fft = new FFT2();
                _realSignal = _fourierSignal.copy();
                fft.inverseTransform(_realSignal, FFTDemo.this);
                _normalImage.setImage(FourierTransform2D.realToGrayscale(_realSignal));
                updateTransformButtons();
            }
        };
        t.start();
    }

    private void invokeFFT() {
        __log.info("Invoking forward fft");
        final TransformThread t = new TransformThread(this, "Radix2 FFT") {
            public void transform() {
                FFT2 fft = new FFT2();
                final BufferedImage image = (BufferedImage) _normalImage.getImage();
                ComplexSignal signal = FourierTransform2D.imageToSignal(image);
                if (!FourierTransform2D.isPowerOf2(new Dimension(signal.width, signal.height)))
                    signal = fft.createPowerOf2Image(signal);

                fft.transform(signal, FFTDemo.this);
                _freqImage.setImage(FourierTransform2D.fourierSignalToGrayscale(signal));
                _fourierSignal = signal;
                updateTransformButtons();
            }
        };
        t.start();
    }

    private void invokeInverseDFT(){
        __log.info("Invoking inverse dft");
        final TransformThread t = new TransformThread(this,"Inverse naive FFT") {
            public void transform() {
                final DFT2 dft = new DFT2();
                _realSignal = _fourierSignal.copy();
                dft.inverseTransform(_realSignal, FFTDemo.this);
                _normalImage.setImage(FourierTransform2D.realToGrayscale(_realSignal));
                updateTransformButtons();
            }
        };
        t.start();
    }

    private void invokeDFT() {
        __log.info("Invoking forward dft");
        final TransformThread t = new TransformThread(this, "Naive DFT") {
            public void transform() {
                DFT2 fft = new DFT2();
                final BufferedImage image = (BufferedImage) _normalImage.getImage();
                ComplexSignal signal = FourierTransform2D.imageToSignal(image);
                fft.transform(signal, FFTDemo.this);

                _freqImage.setImage(FourierTransform2D.fourierSignalToGrayscale(signal));
                _fourierSignal = signal;
                updateTransformButtons();
            }
        };
        t.start();
    }

    private void initSharedFileChooser() {
        _sharedFileChooser = new JFileChooser(".");
        _imageFilter = new FileFilter() {
            public boolean accept(File f) {
                final String name = f.getName().toLowerCase();
                return f.isDirectory() || name.endsWith(".png") || name.endsWith(".jpg") || name.endsWith(".gif");
            }

            public String getDescription() {
                return "Image files .png .jpg .gif";
            }
        };
        _complexFilter = new FileFilter() {
            public boolean accept(File f) {
                final String name = f.getName().toLowerCase();
                return f.isDirectory() || name.endsWith(".cid");
            }

            public String getDescription() {
                return "Complex image data .cid";
            }
        };
        _sharedFileChooser.addChoosableFileFilter(_imageFilter);
        _sharedFileChooser.addChoosableFileFilter(_complexFilter);
    }

    private void openNormalImage() {
        _sharedFileChooser.removeChoosableFileFilter(_imageFilter);
        _sharedFileChooser.removeChoosableFileFilter(_complexFilter);
        _sharedFileChooser.addChoosableFileFilter(_imageFilter);

        BufferedImage input = null;

        _sharedFileChooser.setDialogTitle("Import image...");
        if (_sharedFileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
            final File f = _sharedFileChooser.getSelectedFile();
            try {
                input = ImageIO.read(f);
            } catch (IOException e) {
                JOptionPane.showMessageDialog(this, "The specified image\ncould not be read.", "Error", JOptionPane.ERROR_MESSAGE);
            }
            if (input != null) {
                final BufferedImage image = new BufferedImage(input.getWidth(), input.getHeight(), BufferedImage.TYPE_BYTE_GRAY);
                image.createGraphics().drawImage(input, 0, 0, null);
                _realSignal = FourierTransform2D.imageToSignal(image);
                _normalImage.setImage(image);
                _statusTF.setText("Image imported successfully.");
            } else
                JOptionPane.showMessageDialog(null,"The image could not be imported.\nThe file seems to be invalid.","Error",JOptionPane.ERROR_MESSAGE);

        }
        updateTransformButtons();
    }

    public static void main(String[] args) {
        new FFTDemo();
    }

    private void openFrequencyImage(){
        _sharedFileChooser.removeChoosableFileFilter(_imageFilter);
        _sharedFileChooser.removeChoosableFileFilter(_complexFilter);
        _sharedFileChooser.addChoosableFileFilter(_complexFilter);

        if (_sharedFileChooser.showOpenDialog(null) == JFileChooser.APPROVE_OPTION) {
            try {
                final ObjectInputStream in = new ObjectInputStream(new BufferedInputStream(
                    new FileInputStream(_sharedFileChooser.getSelectedFile())));
                final ComplexSignal s = (ComplexSignal) in.readObject();
                in.close();

                _fourierSignal = s;
                _freqImage.setImage(FourierTransform2D.fourierSignalToGrayscale(_fourierSignal));
                __log.info("signals equal: "+s.equals(_fourierSignal));
            } catch (Exception e) {
                __log.error("Could not read complex data file: ",e);
                JOptionPane.showMessageDialog(null,"Could not import the specified complex image","Error",
                    JOptionPane.ERROR_MESSAGE);
            }
        }

        updateTransformButtons();
    }

    private void saveFrequencyImage(){
        if (_fourierSignal == null)
            return;

        _sharedFileChooser.removeChoosableFileFilter(_imageFilter);
        _sharedFileChooser.removeChoosableFileFilter(_complexFilter);
        _sharedFileChooser.addChoosableFileFilter(_complexFilter);

        if (_sharedFileChooser.showSaveDialog(null) == JFileChooser.APPROVE_OPTION) {
            File selection = _sharedFileChooser.getSelectedFile();
            if (!selection.getName().toLowerCase().endsWith(".cid")){
                final String name = selection.getName()+".cid";
                selection = new File(selection.getParentFile(),name);
            }

            try {
                if (!selection.exists())
                    selection.createNewFile();

                final ObjectOutputStream oos = new ObjectOutputStream(
                    new BufferedOutputStream(new FileOutputStream(selection)));
                oos.writeObject(_fourierSignal);
                oos.flush();
                oos.close();
                _statusTF.setText("Complex data exported successfully.");
            } catch (IOException e) {
                __log.error("Could not export frequency image",e);
                JOptionPane.showMessageDialog(null,"Frequency image could not be saved. See the log for details.",
                    "Error",JOptionPane.ERROR_MESSAGE);
            }
        }
    }

    public void setProgress(final String activity, final float percentage) {
        _progressBar.setString(activity);
        _progressBar.setValue((int) (percentage*100));
        __log.info("Progress: "+percentage);
    }

    public void updateTransformButtons(){
        if (_realSignal != null) {
            _dftButton.setEnabled(true);
            _fftButton.setEnabled(FourierTransform2D.isPowerOf2(_realSignal.width) &&
                FourierTransform2D.isPowerOf2(_realSignal.height));
        } else {
            _dftButton.setEnabled(false);
            _fftButton.setEnabled(false);
        }
        if (_fourierSignal != null) {
            _iDftButton.setEnabled(true);
            _iFftButton.setEnabled(FourierTransform2D.isPowerOf2(_fourierSignal.width) &&
                FourierTransform2D.isPowerOf2(_fourierSignal.height));
        } else {
            _iDftButton.setEnabled(false);
            _iFftButton.setEnabled(false);
        }
    }
}
